package game.biome;

import java.util.List;
import java.util.Map;

import game.block.Block;
import game.block.BlockColored;
import game.block.BlockFlower;
import game.block.BlockSand;
import game.block.BlockTallGrass;
import game.collect.Lists;
import game.collect.Maps;
import game.color.Colorizer;
import game.color.DyeColor;
import game.entity.animal.EntityBat;
import game.entity.animal.EntityChicken;
import game.entity.animal.EntityCow;
import game.entity.animal.EntityMouse;
import game.entity.animal.EntityPig;
import game.entity.animal.EntityRabbit;
import game.entity.animal.EntitySheep;
import game.entity.animal.EntitySquid;
import game.entity.npc.EntityArachnoid;
import game.entity.npc.EntityHaunter;
import game.entity.npc.EntityMage;
import game.entity.npc.EntitySlime;
import game.entity.npc.EntityUndead;
import game.entity.npc.EntityZombie;
import game.entity.types.EntityLiving;
import game.init.Blocks;
import game.log.Log;
import game.material.Material;
import game.rng.PerlinGen;
import game.rng.Random;
import game.rng.RngItem;
import game.rng.WeightedList;
import game.util.ExtMath;
import game.world.BlockPos;
import game.world.State;
import game.world.World;
import game.world.WorldServer;
import game.worldgen.ChunkPrimer;
import game.worldgen.FeatureGenerator;
import game.worldgen.feature.WorldGenClay;
import game.worldgen.feature.WorldGenClayExt;
import game.worldgen.feature.WorldGenSand;
import game.worldgen.foliage.FeatureDoublePlant;
import game.worldgen.foliage.WorldGenBigMushroom;
import game.worldgen.foliage.WorldGenCactus;
import game.worldgen.foliage.WorldGenDeadBush;
import game.worldgen.foliage.WorldGenFlowers;
import game.worldgen.foliage.WorldGenMushroom;
import game.worldgen.foliage.WorldGenPumpkin;
import game.worldgen.foliage.WorldGenReed;
import game.worldgen.foliage.WorldGenTallGrass;
import game.worldgen.foliage.WorldGenWaterlily;
import game.worldgen.tree.WorldGenBaseTree;
import game.worldgen.tree.WorldGenBigTree;
import game.worldgen.tree.WorldGenSwamp;
import game.worldgen.tree.WorldGenTree;

public abstract class Biome
{
    protected static final Biome.Height height_Default = new Biome.Height(0.1F, 0.2F);
    protected static final Biome.Height height_ShallowWaters = new Biome.Height(-0.5F, 0.0F);
    protected static final Biome.Height height_Sea = new Biome.Height(-1.0F, 0.1F);
    protected static final Biome.Height height_DeepSea = new Biome.Height(-1.8F, 0.1F);
    protected static final Biome.Height height_LowPlains = new Biome.Height(0.125F, 0.05F);
    protected static final Biome.Height height_MidPlains = new Biome.Height(0.2F, 0.2F);
    protected static final Biome.Height height_LowHills = new Biome.Height(0.45F, 0.3F);
    protected static final Biome.Height height_HighPlateaus = new Biome.Height(1.5F, 0.025F);
    protected static final Biome.Height height_MidHills = new Biome.Height(1.0F, 0.5F);
    protected static final Biome.Height height_Shores = new Biome.Height(0.0F, 0.025F);
    protected static final Biome.Height height_RockyWaters = new Biome.Height(0.1F, 0.8F);
    protected static final Biome.Height height_MidRockyPlains = new Biome.Height(0.2F, 0.3F);
    protected static final Biome.Height height_PartiallySubmerged = new Biome.Height(-0.2F, 0.1F);
    protected static final Biome.Height height_Varying = new Biome.Height(0.1F, 1.0F);
    protected static final Biome.Height height_VaryingExtreme = new Biome.Height(1.0F, 2.0F);
    
    private static final Biome[] BIOMES = new Biome[256];
    public static final Map<String, Biome> BIOME_ID_MAP = Maps.<String, Biome>newHashMap();
    
    public static final Biome none = (new BiomeNone(0)).setBiomeName("none", "<Keins>");
    
    public static final Biome plains = (new BiomePlains(1)).setColor(9286496).setBiomeName("plains", "Ebene");
    public static final Biome desert = (new BiomeDesert(2)).setColor(16421912).setBiomeName("desert", "Wüste").setTemperatureLegacy(2.0F).setHumidity(0.0F).setHeight(height_LowPlains);
    public static final Biome extremeHills = (new BiomeHills(3, false)).setColor(6316128).setBiomeName("extremeHills", "Extremes Bergland").setHeight(height_MidHills).setTemperatureLegacy(0.2F).setHumidity(0.3F);
    public static final Biome forest = (new BiomeForest(4, 0)).setColor(353825).setBiomeName("forest", "Wald");
    public static final Biome taiga = (new BiomeTaiga(5, 0)).setColor(747097).setBiomeName("taiga", "Taiga").setTemperatureLegacy(0.25F).setHumidity(0.8F).setHeight(height_MidPlains);
    public static final Biome swampland = (new BiomeSwamp(6)).setColor(522674).setBiomeName("swampland", "Sumpf").setHeight(height_PartiallySubmerged).setTemperatureLegacy(0.8F).setHumidity(0.9F);
    public static final Biome river = (new BiomeWater(7, true)).setColor(255).setBiomeName("river", "Fluss").setHeight(height_ShallowWaters);
    
    public static final Biome exterminated = (new BiomeExterminated(8)).setColor(0x000000).setBiomeName("exterminated", "Ausgelöscht").setHumidity(0.0F).setTemperature(150.0f);
    public static final Biome space = (new BiomeSpace(9)).setColor(0x000000).setBiomeName("space", "Leere des Weltraums").setHumidity(0.0F);
    
    public static final Biome frozenSea = (new BiomeWater(10, false)).setColor(9474208).setBiomeName("frozenSea", "Vereister See").setSnowy().setHeight(height_Sea).setTemperatureLegacy(0.0F).setHumidity(0.5F);
    public static final Biome frozenRiver = (new BiomeWater(11, true)).setColor(10526975).setBiomeName("frozenRiver", "Vereister Fluss").setSnowy().setHeight(height_ShallowWaters).setTemperatureLegacy(0.0F).setHumidity(0.5F);
    public static final Biome icePlains = (new BiomeSnow(12, false)).setColor(16777215).setBiomeName("icePlains", "Eisebene").setSnowy().setTemperatureLegacy(0.0F).setHumidity(0.5F).setHeight(height_LowPlains);
    public static final Biome iceMountains = (new BiomeSnow(13, false)).setColor(10526880).setBiomeName("iceMountains", "Vereistes Bergland").setSnowy().setHeight(height_LowHills).setTemperatureLegacy(0.0F).setHumidity(0.5F);
    public static final Biome mushroomPlains = (new BiomeMushroom(14)).setColor(16711935).setBiomeName("mushroomPlains", "Pilzland").setTemperatureLegacy(0.9F).setHumidity(1.0F).setHeight(height_MidRockyPlains);
//    public static final Biome mushroomPlainsEdge = (new BiomeMushroom(15)).setColor(10486015).setBiomeName("mushroomPlainsEdge").setTemperatureLegacy(0.9F).setHumidity(1.0F).setHeight(height_Shores);
    public static final Biome beach = (new BiomeBeach(16)).setColor(16440917).setBiomeName("beach", "Strand").setTemperatureLegacy(0.8F).setHumidity(0.4F).setHeight(height_Shores);
    public static final Biome desertHills = (new BiomeDesert(17)).setColor(13786898).setBiomeName("desertHills", "Wüsten-Bergland").setTemperatureLegacy(2.0F).setHumidity(0.0F).setHeight(height_LowHills);
    public static final Biome forestHills = (new BiomeForest(18, 0)).setColor(2250012).setBiomeName("forestHills", "Wald-Bergland").setHeight(height_LowHills);
    public static final Biome taigaHills = (new BiomeTaiga(19, 0)).setColor(1456435).setBiomeName("taigaHills", "Taiga-Bergland").setTemperatureLegacy(0.25F).setHumidity(0.8F).setHeight(height_LowHills);
    public static final Biome extremeHillsEdge = (new BiomeHills(20, true)).setColor(7501978).setBiomeName("extremeHillsEdge", "Extremes Bergland Gr.").setHeight(height_MidHills.attenuate()).setTemperatureLegacy(0.2F).setHumidity(0.3F);
    public static final Biome jungle = (new BiomeJungle(21, false)).setColor(5470985).setBiomeName("jungle", "Urwald").setTemperatureLegacy(0.95F).setHumidity(0.9F);
    public static final Biome jungleHills = (new BiomeJungle(22, false)).setColor(2900485).setBiomeName("jungleHills", "Urwald-Bergland").setTemperatureLegacy(0.95F).setHumidity(0.9F).setHeight(height_LowHills);
    public static final Biome jungleEdge = (new BiomeJungle(23, true)).setColor(6458135).setBiomeName("jungleEdge", "Urwald Gr.").setTemperatureLegacy(0.95F).setHumidity(0.8F);
    public static final Biome sea = (new BiomeWater(24, false)).setColor(112).setBiomeName("sea", "See").setHeight(height_Sea);
    public static final Biome stoneBeach = (new BiomeStoneBeach(25)).setColor(10658436).setBiomeName("stoneBeach", "Steinstrand").setTemperatureLegacy(0.2F).setHumidity(0.3F).setHeight(height_RockyWaters);
    public static final Biome coldBeach = (new BiomeBeach(26)).setColor(16445632).setBiomeName("coldBeach", "Vereister Strand").setTemperatureLegacy(0.05F).setHumidity(0.3F).setHeight(height_Shores).setSnowy();
    public static final Biome birchForest = (new BiomeForest(27, 2)).setBiomeName("birchForest", "Birkenwald").setColor(3175492);
    public static final Biome birchForestHills = (new BiomeForest(28, 2)).setBiomeName("birchForestHills", "Birkenwald-Bergland").setColor(2055986).setHeight(height_LowHills);
    public static final Biome roofedForest = (new BiomeForest(29, 3)).setColor(4215066).setBiomeName("roofedForest", "Dichter Wald");
    public static final Biome coldTaiga = (new BiomeTaiga(30, 0)).setColor(3233098).setBiomeName("coldTaiga", "Vereiste Taiga").setSnowy().setTemperatureLegacy(-0.5F).setHumidity(0.4F).setHeight(height_MidPlains);
    public static final Biome coldTaigaHills = (new BiomeTaiga(31, 0)).setColor(2375478).setBiomeName("coldTaigaHills", "Vereistes Taiga-Bergland").setSnowy().setTemperatureLegacy(-0.5F).setHumidity(0.4F).setHeight(height_LowHills);
    public static final Biome megaTaiga = (new BiomeTaiga(32, 1)).setColor(5858897).setBiomeName("megaTaiga", "Hohe Taiga").setTemperatureLegacy(0.3F).setHumidity(0.8F).setHeight(height_MidPlains);
    public static final Biome megaTaigaHills = (new BiomeTaiga(33, 1)).setColor(4542270).setBiomeName("megaTaigaHills", "Hohes Taiga-Bergland").setTemperatureLegacy(0.3F).setHumidity(0.8F).setHeight(height_LowHills);
    public static final Biome extremeHillsPlus = (new BiomeHills(34, true)).setColor(5271632).setBiomeName("extremeHillsPlus", "Extremes Bergland +").setHeight(height_MidHills).setTemperatureLegacy(0.2F).setHumidity(0.3F);
    public static final Biome savanna = (new BiomeSavanna(35)).setColor(12431967).setBiomeName("savanna", "Savanne").setTemperature(28.0F).setHumidity(0.0F).setHeight(height_LowPlains);
    public static final Biome savannaPlateau = (new BiomeSavanna(36)).setColor(10984804).setBiomeName("savannaPlateau", "Savannen-Plateau").setTemperature(20.0F).setHumidity(0.0F).setHeight(height_HighPlateaus);
    
    public static final Biome mesa = (new BiomeMesa(37, false, false)).setColor(14238997).setBiomeName("mesa", "Mesa");
    public static final Biome mesaPlateau_F = (new BiomeMesa(38, false, true)).setColor(11573093).setBiomeName("mesaPlateauF", "Mesa-Waldplateau").setHeight(height_HighPlateaus);
    public static final Biome mesaPlateau = (new BiomeMesa(39, false, false)).setColor(13274213).setBiomeName("mesaPlateau", "Mesa-Plateau").setHeight(height_HighPlateaus);
    
    public static final Biome snowLand = (new BiomeSnowLand(40)).setColor(0xffffff).setBiomeName("snowLand", "Eisland").setSnowy().setHumidity(1.0F);
    public static final Biome tian = (new BiomeTian(41)).setColor(0x808080).setBiomeName("tian", "Tian").setHumidity(0.8F).setHeight(height_Varying);
    public static final Biome elvenForest = (new BiomeForest(42, 4)).setColor(0x059821).setBiomeName("elvenForest", "Elbenwald").setHumidity(0.9F);
    public static final Biome upperHell = (new BiomeHell(43, 0)).setColor(16711680).setBiomeName("upperHell", "Übergang in die Hölle").setHumidity(0.0F);
    public static final Biome lowerHell = (new BiomeHell(44, 1)).setColor(16711680).setBiomeName("lowerHell", "Abgrund der Hölle").setHumidity(0.0F);
    public static final Biome hellHills = (new BiomeHell(45, 1)).setColor(16711680).setBiomeName("hellHills", "Bergland der Hölle").setHumidity(0.0F).setHeight(height_MidHills);
    public static final Biome soulPlains = (new BiomeHell(46, 1)).setColor(16711680).setBiomeName("soulPlains", "Seelenland").setHumidity(0.0F).setHeight(height_PartiallySubmerged);
    public static final Biome ashLand = (new BiomeHell(47, 2)).setColor(16711680).setBiomeName("ashLand", "Verbrannt").setHumidity(0.0F).setHeight(height_LowPlains);
//    public static final Biome star = (new BiomeStar(46)).setColor(0x000000).setBiomeName("star").setHumidity(0.0F);
    public static final Biome moon = (new BiomeMoon(48)).setColor(0xa0a0a0).setBiomeName("moon", "Mondoberfläche").setHumidity(0.0F).setHeight(height_LowPlains);
    public static final Biome chaos = (new BiomeChaos(49)).setColor(0xff00ff).setBiomeName("chaos", "Chaos").setHumidity(0.5F).setHeight(height_VaryingExtreme);
//    public static final Biome magmaLands = (new BiomeHell(49, 1)).setColor(16711680).setBiomeName("magmaLands").setRainfall(0.0F).setHeight(height_PartiallySubmerged);
    
    public static final Biome DEF_BIOME = forest;
    protected static final PerlinGen TEMP_NOISE;
    protected static final PerlinGen TREE_NOISE;
    protected static final PerlinGen GRASS_NOISE;
    protected static final FeatureDoublePlant DOUBLE_PLANT_GEN;
    private static final Map<String, Biome> LOOKUP = Maps.newTreeMap();
    private static final List<Biome> LIST = Lists.newArrayList();
    
    public final int id;
    
    protected final WeightedList<Biome.RngSpawn> mobs = new WeightedList<Biome.RngSpawn>();
    protected final WorldGenBaseTree worldGeneratorTrees = new WorldGenBaseTree(false);
    protected final WorldGenBigTree worldGeneratorBigTree = new WorldGenBigTree(false);
    protected final WorldGenSwamp worldGeneratorSwamp = new WorldGenSwamp();
    private final FeatureGenerator clayGen = new WorldGenClay(4);
    private final FeatureGenerator sandGen = new WorldGenSand(Blocks.sand, 7);
    private final FeatureGenerator gravelAsSandGen = new WorldGenSand(Blocks.gravel, 6);
    private final WorldGenFlowers yellowFlowerGen = new WorldGenFlowers(Blocks.flower, BlockFlower.EnumFlowerType.DANDELION);
    private final FeatureGenerator mushroomBrownGen = new WorldGenMushroom(Blocks.brown_mushroom);
    private final FeatureGenerator mushroomRedGen = new WorldGenMushroom(Blocks.red_mushroom);
    private final FeatureGenerator bigMushroomGen = new WorldGenBigMushroom();
    private final FeatureGenerator reedGen = new WorldGenReed();
    private final FeatureGenerator cactusGen = new WorldGenCactus();
    private final FeatureGenerator waterlilyGen = new WorldGenWaterlily();
    private final FeatureGenerator clayGenExt = new WorldGenClayExt(32);
    
    public String name = null;
    public String display = "<?>";
    public int color = 0x000000;
    public State topBlock = Blocks.grass.getState();
    public State fillerBlock = Blocks.dirt.getState();
    public float minHeight = height_Default.rootHeight;
    public float maxHeight = height_Default.variation;
    public float temperature = 0.0F;
    public float humidity = 0.5F;
    public int waterColor = 0xffffff;
    public boolean generateLakes = true;
    public boolean generateLiquids = true;
    public boolean snowyGen = false;
    public boolean waterGen = false;

    protected int waterlilyPerChunk = 0;
    protected int treesPerChunk = 0;
    protected int flowersPerChunk = 2;
    protected int grassPerChunk = 1;
    protected int deadBushPerChunk = 0;
    protected int mushroomsPerChunk = 0;
    protected int reedsPerChunk = 0;
    protected int cactiPerChunk = 0;
    protected int sandPerChunk = 1;
    protected int sandPerChunk2 = 3;
    protected int clayPerChunk = 1;
    protected int clayExtPerChunk = 0; // 10
    protected int bigMushroomsPerChunk = 0;

    protected Biome(int id) {
        this.id = id;
        BIOMES[id] = this;
        this.addMobs(this.mobs);
    }
    
    protected void addMobs(WeightedList<Biome.RngSpawn> mobs) {
    	mobs.add(new Biome.RngSpawn(EntitySheep.class, 12, 4, 4));
        mobs.add(new Biome.RngSpawn(EntityRabbit.class, 10, 3, 10));
        mobs.add(new Biome.RngSpawn(EntityPig.class, 10, 4, 4));
        mobs.add(new Biome.RngSpawn(EntityChicken.class, 10, 4, 4));
        mobs.add(new Biome.RngSpawn(EntityCow.class, 8, 4, 4));
        mobs.add(new Biome.RngSpawn(EntityArachnoid.class, 100, 4, 4));
        mobs.add(new Biome.RngSpawn(EntityZombie.class, 100, 4, 4));
        mobs.add(new Biome.RngSpawn(EntityUndead.class, 100, 4, 4));
        mobs.add(new Biome.RngSpawn(EntityHaunter.class, 100, 4, 4));
        mobs.add(new Biome.RngSpawn(EntitySlime.class, 100, 4, 4));
//        mobs.add(new Biome.RngSpawn(EntityEnder....class, 10, 1, 4));
        mobs.add(new Biome.RngSpawn(EntityMage.class, 5, 1, 1));
        mobs.add(new Biome.RngSpawn(EntitySquid.class, 10, 4, 4));
        mobs.add(new Biome.RngSpawn(EntityBat.class, 10, 8, 8));
        mobs.add(new Biome.RngSpawn(EntityMouse.class, 10, 8, 8));
    }
    
    public int getSkyColor() {
    	return 0xffffffff;
    }
    
    public int getFogColor() {
    	return 0xffffffff;
    }
    
    public int getCloudColor() {
    	return 0xffffffff;
    }
    
    protected Biome setTemperatureLegacy(float temperatureIn) // TODO: away, begone!!!
    {
        return this.setTemperature(((temperatureIn - 0.15f) * 40.0f) - 14.0f);
    }
    
    protected Biome setTemperature(float temperatureIn)
    {
        this.temperature = temperatureIn;
        return this;
    }
    
    protected Biome setHumidity(float humidity)
    {
        this.humidity = humidity;
        return this;
    }

    protected final Biome setHeight(Biome.Height heights)
    {
        this.minHeight = heights.rootHeight;
        this.maxHeight = heights.variation;
        return this;
    }

    public WorldGenTree genBigTreeChance(Random rand)
    {
        return rand.chance(10) ? this.worldGeneratorBigTree : this.worldGeneratorTrees;
    }

    public WorldGenTree genBigTreeLegacy(Random rand, BlockPos pos)
    {
        int noise = (int)((TREE_NOISE.generate((double)pos.getX() * 0.5D, (double)pos.getZ() * 0.5D) / 8D + rand.doublev() * 4D + 4D) / 3D);
        return (noise > 0 && rand.chance(noise)) || (this.isHighHumidity() && rand.chance(3)) ? this.worldGeneratorBigTree : 
        		this.worldGeneratorTrees;
    }

    /**
     * Gets a WorldGen appropriate for this biome.
     */
    public FeatureGenerator getRandomWorldGenForGrass(Random rand)
    {
        return new WorldGenTallGrass(BlockTallGrass.EnumType.GRASS);
    }

    public BlockFlower.EnumFlowerType pickRandomFlower(Random rand, BlockPos pos)
    {
        return rand.rarity(3) ? BlockFlower.EnumFlowerType.DANDELION : BlockFlower.EnumFlowerType.ROSE;
    }
    
    protected Biome setSnowy()
    {
        this.snowyGen = true;
        return this;
    }
    
    protected Biome setWater()
    {
        this.waterGen = true;
        return this;
    }

    protected Biome setBiomeName(String name, String display)
    {
        this.name = name;
        this.display = display;
        return this;
    }

    protected Biome setColor(int colorIn)
    {
    	this.color = colorIn;
        return this;
    }

    public WeightedList<Biome.RngSpawn> getSpawnableList()
    {
        return this.mobs;
    }
    
    public boolean isHighHumidity()
    {
        return this.humidity > 0.85F;
    }

    /**
     * returns the chance a creature has to spawn.
     */
    public float getSpawningChance()
    {
        return 0.1F;
    }

    /**
     * Gets a floating point representation of this biome's temperature
     */
    public final float getFloatTemperature(BlockPos pos)
    {
        if (pos.getY() > 64)
        {
            float f = (float)(TEMP_NOISE.generate((double)pos.getX() * 1.0D / 8.0D, (double)pos.getZ() * 1.0D / 8.0D) * 4.0D);
            return this.temperature - (f + (float)pos.getY() - 64.0F) / 15.0f;
        }
        else
        {
            return this.temperature;
        }
    }
    
    public void decorate(WorldServer world, Random rand, BlockPos pos)
    {
        for (int i = 0; i < this.sandPerChunk2; ++i)
        {
            int j = rand.chOffset();
            int k = rand.chOffset();
            this.sandGen.generate(world, rand, world.getTopSolidOrLiquidBlock(pos.add(j, 0, k)));
        }

        for (int i1 = 0; i1 < this.clayPerChunk; ++i1)
        {
            int l1 = rand.chOffset();
            int i6 = rand.chOffset();
            this.clayGen.generate(world, rand, world.getTopSolidOrLiquidBlock(pos.add(l1, 0, i6)));
        }

        for (int j1 = 0; j1 < this.sandPerChunk; ++j1)
        {
            int i2 = rand.chOffset();
            int j6 = rand.chOffset();
            this.gravelAsSandGen.generate(world, rand, world.getTopSolidOrLiquidBlock(pos.add(i2, 0, j6)));
        }

        for (int i1 = 0; i1 < this.clayExtPerChunk; ++i1)
        {
            int l1 = rand.chOffset();
            int i6 = rand.chOffset();
            this.clayGenExt.generate(world, rand, world.getTopSolidOrLiquidBlock(pos.add(l1, 0, i6)));
        }

        int k1 = this.treesPerChunk;

        if (rand.chance(10))
        {
            ++k1;
        }

        for (int j2 = 0; j2 < k1; ++j2)
        {
            int k6 = rand.chOffset();
            int l = rand.chOffset();
            WorldGenTree treeGen = this.genBigTreeChance(rand);
            treeGen.prepare();
            BlockPos blockpos = world.getHeight(pos.add(k6, 0, l));

            if (treeGen.generate(world, rand, blockpos))
            {
                treeGen.finish(world, rand, blockpos);
            }
        }

        for (int k2 = 0; k2 < this.bigMushroomsPerChunk; ++k2)
        {
            int l6 = rand.chOffset();
            int k10 = rand.chOffset();
            this.bigMushroomGen.generate(world, rand, world.getHeight(pos.add(l6, 0, k10)));
        }

        for (int l2 = 0; l2 < this.flowersPerChunk; ++l2)
        {
            int i7 = rand.chOffset();
            int l10 = rand.chOffset();
            int j14 = world.getHeight(pos.add(i7, 0, l10)).getY() + 32;

            if (j14 > 0)
            {
                int k17 = rand.zrange(j14);
                BlockPos blockpos1 = pos.add(i7, k17, l10);
                BlockFlower.EnumFlowerType blockflower$enumflowertype = this.pickRandomFlower(rand, blockpos1);
                BlockFlower blockflower = blockflower$enumflowertype.getBlockType().getBlock();

                if (blockflower.getMaterial() != Material.air)
                {
                    this.yellowFlowerGen.setGeneratedBlock(blockflower, blockflower$enumflowertype);
                    this.yellowFlowerGen.generate(world, rand, blockpos1);
                }
            }
        }

        for (int i3 = 0; i3 < this.grassPerChunk; ++i3)
        {
            int j7 = rand.chOffset();
            int i11 = rand.chOffset();
            int k14 = world.getHeight(pos.add(j7, 0, i11)).getY() * 2;

            if (k14 > 0)
            {
                int l17 = rand.zrange(k14);
                this.getRandomWorldGenForGrass(rand).generate(world, rand, pos.add(j7, l17, i11));
            }
        }

        for (int j3 = 0; j3 < this.deadBushPerChunk; ++j3)
        {
            int k7 = rand.chOffset();
            int j11 = rand.chOffset();
            int l14 = world.getHeight(pos.add(k7, 0, j11)).getY() * 2;

            if (l14 > 0)
            {
                int i18 = rand.zrange(l14);
                (new WorldGenDeadBush()).generate(world, rand, pos.add(k7, i18, j11));
            }
        }

        for (int k3 = 0; k3 < this.waterlilyPerChunk; ++k3)
        {
            int l7 = rand.chOffset();
            int k11 = rand.chOffset();
            int i15 = world.getHeight(pos.add(l7, 0, k11)).getY() * 2;

            if (i15 > 0)
            {
                int j18 = rand.zrange(i15);
                BlockPos blockpos4;
                BlockPos blockpos7;

                for (blockpos4 = pos.add(l7, j18, k11); blockpos4.getY() > 0; blockpos4 = blockpos7)
                {
                    blockpos7 = blockpos4.down();

                    if (!world.isAirBlock(blockpos7))
                    {
                        break;
                    }
                }

                this.waterlilyGen.generate(world, rand, blockpos4);
            }
        }

        for (int l3 = 0; l3 < this.mushroomsPerChunk; ++l3)
        {
            if (rand.chance(4))
            {
                int i8 = rand.chOffset();
                int l11 = rand.chOffset();
                BlockPos blockpos2 = world.getHeight(pos.add(i8, 0, l11));
                this.mushroomBrownGen.generate(world, rand, blockpos2);
            }

            if (rand.chance(8))
            {
                int j8 = rand.chOffset();
                int i12 = rand.chOffset();
                int j15 = world.getHeight(pos.add(j8, 0, i12)).getY() * 2;

                if (j15 > 0)
                {
                    int k18 = rand.zrange(j15);
                    BlockPos blockpos5 = pos.add(j8, k18, i12);
                    this.mushroomRedGen.generate(world, rand, blockpos5);
                }
            }
        }

        if (this.mushroomsPerChunk != -1 && rand.chance(4))
        {
            int i4 = rand.chOffset();
            int k8 = rand.chOffset();
            int j12 = world.getHeight(pos.add(i4, 0, k8)).getY() * 2;

            if (j12 > 0)
            {
                int k15 = rand.zrange(j12);
                this.mushroomBrownGen.generate(world, rand, pos.add(i4, k15, k8));
            }
        }

        if (this.mushroomsPerChunk != -1 && rand.chance(8))
        {
            int j4 = rand.chOffset();
            int l8 = rand.chOffset();
            int k12 = world.getHeight(pos.add(j4, 0, l8)).getY() * 2;

            if (k12 > 0)
            {
                int l15 = rand.zrange(k12);
                this.mushroomRedGen.generate(world, rand, pos.add(j4, l15, l8));
            }
        }

        for (int k4 = 0; k4 < this.reedsPerChunk; ++k4)
        {
            int i9 = rand.chOffset();
            int l12 = rand.chOffset();
            int i16 = world.getHeight(pos.add(i9, 0, l12)).getY() * 2;

            if (i16 > 0)
            {
                int l18 = rand.zrange(i16);
                this.reedGen.generate(world, rand, pos.add(i9, l18, l12));
            }
        }

        for (int l4 = 0; l4 < 10; ++l4)
        {
            int j9 = rand.chOffset();
            int i13 = rand.chOffset();
            int j16 = world.getHeight(pos.add(j9, 0, i13)).getY() * 2;

            if (j16 > 0)
            {
                int i19 = rand.zrange(j16);
                this.reedGen.generate(world, rand, pos.add(j9, i19, i13));
            }
        }

        if (rand.chance(32))
        {
            int i5 = rand.chOffset();
            int k9 = rand.chOffset();
            int j13 = world.getHeight(pos.add(i5, 0, k9)).getY() * 2;

            if (j13 > 0)
            {
                int k16 = rand.zrange(j13);
                (new WorldGenPumpkin()).generate(world, rand, pos.add(i5, k16, k9));
            }
        }

        for (int j5 = 0; j5 < this.cactiPerChunk; ++j5)
        {
            int l9 = rand.chOffset();
            int k13 = rand.chOffset();
            int l16 = world.getHeight(pos.add(l9, 0, k13)).getY() * 2;

            if (l16 > 0)
            {
                int j19 = rand.zrange(l16);
                this.cactusGen.generate(world, rand, pos.add(l9, j19, k13));
            }
        }
    }

    public int getGrassColorAtPos(BlockPos pos)
    {
        double d0 = (double)ExtMath.clampf((this.getFloatTemperature(pos) + 14.0f) / 40.0f + 0.15f, 0.0F, 1.0F);
        double d1 = (double)ExtMath.clampf(this.humidity, 0.0F, 1.0F);
        return Colorizer.getGrassColor(d0, d1);
    }

    public int getFoliageColorAtPos(BlockPos pos)
    {
        double d0 = (double)ExtMath.clampf((this.getFloatTemperature(pos) + 14.0f) / 40.0f + 0.15f, 0.0F, 1.0F);
        double d1 = (double)ExtMath.clampf(this.humidity, 0.0F, 1.0F);
        return Colorizer.getFoliageColor(d0, d1);
    }

    public boolean isSnowyBiome()
    {
        return this.snowyGen;
    }

    public void genTerrainBlocks(WorldServer worldIn, Random rand, ChunkPrimer chunkPrimerIn, int x, int z, double noiseVal)
    {
        this.generateBiomeTerrain(worldIn, rand, chunkPrimerIn, x, z, noiseVal);
    }

    /**
     * Given x, z coordinates, we count down all the y positions starting at height - 1 and working our way down. When we hit a
     * non-air block, we replace it with this.topBlock (default grass, descendants may set otherwise), and then a
     * relatively shallow layer of blocks of type this.fillerBlock (default dirt). A random set of blocks below y == 5
     * (but always including y == 0) is replaced with bedrock in Chunk(...).
     *  
     * If we don't hit non-air until somewhat below sea level, we top with gravel and fill down with stone.
     *  
     * If this.fillerBlock is red sand, we replace some of that with red sandstone.
     */
    public final void generateBiomeTerrain(WorldServer worldIn, Random rand, ChunkPrimer chunkPrimerIn, int x, int z, double noiseVal)
    {
        int i = worldIn.getSeaLevel();
        State worldState = worldIn.dimension.getFiller();
        Block worldBlock = worldState.getBlock();
        State worldAlt = worldIn.dimension.getAltFiller1();
        State liquid = worldIn.getSurfaceLiquid();
        boolean freeze = liquid.getBlock().getMaterial() == Material.water;
        State iblockstate = this.topBlock;
        State iblockstate1 = this.fillerBlock;
        int j = -1;
        int k = (int)(noiseVal / 3.0D + 3.0D + rand.doublev() * 0.25D);
        int l = x & 15;
        int i1 = z & 15;
        BlockPos.MutableBlockPos blockpos$mutableblockpos = new BlockPos.MutableBlockPos();

        for (int j1 = chunkPrimerIn.height - 1; j1 >= 0; --j1)
        {
            State iblockstate2 = chunkPrimerIn.get(i1, j1, l);

            if (iblockstate2.getBlock().getMaterial() == Material.air)
            {
                j = -1;
            }
            else if (iblockstate2.getBlock() == worldBlock)
            {
                if (j == -1)
                {
                    if (k <= 0)
                    {
                        iblockstate = null;
                        iblockstate1 = worldState;
                    }
                    else if (j1 >= i - 4 && j1 <= i + 1)
                    {
                        iblockstate = this.topBlock;
                        iblockstate1 = this.fillerBlock;
                    }

                    if (j1 < i && (iblockstate == null || iblockstate.getBlock().getMaterial() == Material.air))
                    {
                    	if (freeze && World.ABSOLUTE_ZERO + worldIn.getTempOffset() + this.getFloatTemperature(blockpos$mutableblockpos.set(x, j1, z)) <= 0.0F)
                        {
                            iblockstate = Blocks.ice.getState();
                        }
                        else
                        {
                            iblockstate = liquid;
                        }
                    }

                    j = k;

                    if (j1 >= i - 1)
                    {
                        chunkPrimerIn.set(i1, j1, l, iblockstate);
                    }
                    else if (j1 < i - 7 - k)
                    {
                        iblockstate = null;
                        iblockstate1 = worldState;
                        chunkPrimerIn.set(i1, j1, l, worldAlt);
                    }
                    else
                    {
                        chunkPrimerIn.set(i1, j1, l, iblockstate1);
                    }
                }
                else if (j > 0)
                {
                    --j;
                    chunkPrimerIn.set(i1, j1, l, iblockstate1);

                    if (j == 0 && iblockstate1.getBlock() == Blocks.sand)
                    {
                        j = rand.zrange(4) + Math.max(0, j1 - 63);
                        iblockstate1 = iblockstate1.getValue(BlockSand.VARIANT) == BlockSand.EnumType.RED_SAND ? Blocks.stained_hardened_clay.getState().withProperty(BlockColored.COLOR, DyeColor.ORANGE) : Blocks.sandstone.getState(); //TODO: check!
                    }
                }
            }
        }
    }
    
    protected Biome createMutation()
    {
        return this.createMutatedBiome(this.id + 128);
    }

    protected Biome createMutatedBiome(int p_180277_1_)
    {
        return new BiomeMutated(p_180277_1_, this);
    }

    public Class <? extends Biome > getBiomeClass()
    {
        return this.getClass();
    }
    
    public boolean isEqualTo(Biome biome)
    {
        return biome == this ? true : (biome == null ? false : this.getBiomeClass() == biome.getBiomeClass());
    }

    public Biome.TempCategory getTempCategory()
    {
        return (double)this.temperature < -12.0 ? Biome.TempCategory.COLD : ((double)this.temperature < 20.0 ? Biome.TempCategory.MEDIUM : Biome.TempCategory.WARM);
    }

	public static Biome[] getBiomeGenArray()
    {
        return BIOMES;
    }
	
    public static Biome getBiome(int id)
    {
        return getBiomeFromBiomeList(id, null);
    }

    public static Biome getBiomeFromBiomeList(int biomeId, Biome biome)
    {
        if (biomeId >= 0 && biomeId < BIOMES.length)
        {
            Biome biomegenbase = BIOMES[biomeId];
            return biomegenbase == null ? biome : biomegenbase;
        }
        else
        {
        	Log.JNI.warn("Biom-ID ist nicht im Bereich: " + biomeId + ", verwende " + DEF_BIOME.id + " (" + DEF_BIOME.name + ")");
            return DEF_BIOME;
        }
    }
    
    public static List<Biome> getBiomes() {
        return LIST;
    }
    
    public static List<String> getBiomeNames() {
        return Lists.newArrayList(LOOKUP.keySet());
    }

    public static Biome findByName(String name) {
    	Biome biome = LOOKUP.get(name.toLowerCase().replace(" ", "").replace("_", ""));
    	if(biome == null) {
    		int z;
    		try {
    			z = Integer.parseInt(name);
    		}
    		catch(NumberFormatException e) {
    			return null;
    		}
    		return z < 0 || z >= BIOMES.length ? null : BIOMES[z];
    	}
    	return biome;
	}

    static
    {
//        plains.createMutation();
        desert.createMutation();
        forest.createMutation();
        taiga.createMutation();
        swampland.createMutation();
        icePlains.createMutation();
        jungle.createMutation();
        jungleEdge.createMutation();
        coldTaiga.createMutation();
        savanna.createMutation();
        savannaPlateau.createMutation();
        mesa.createMutation();
        mesaPlateau_F.createMutation();
        mesaPlateau.createMutation();
        birchForest.createMutation();
        birchForestHills.createMutation();
        roofedForest.createMutation();
        megaTaiga.createMutation();
        extremeHills.createMutation();
        extremeHillsPlus.createMutation();
        megaTaiga.createMutatedBiome(megaTaigaHills.id + 128).setBiomeName("redwoodTaigaHillsM", "Mammutbaumtaiga");

        for (Biome biomegenbase : BIOMES)
        {
            if (biomegenbase != null)
            {
                if (BIOME_ID_MAP.containsKey(biomegenbase.name))
                {
                    throw new Error("Biom \"" + biomegenbase.name + "\" ist als ID " + ((Biome)BIOME_ID_MAP.get(biomegenbase.name)).id + " und " + biomegenbase.id + " definiert");
                }

                BIOME_ID_MAP.put(biomegenbase.name, biomegenbase);
            }
        }

        TEMP_NOISE = new PerlinGen(new Random(1234L), 1);
        TREE_NOISE = new PerlinGen(new Random(667L), 8);
        GRASS_NOISE = new PerlinGen(new Random(2345L), 1);
        DOUBLE_PLANT_GEN = new FeatureDoublePlant();
        
        for(Biome biome : BIOMES) {
            if(biome == null) {
                continue;
            }
            LOOKUP.put(biome.name.toLowerCase(), biome);
            LIST.add(biome);
        }
    }

    public static class Height
    {
        public float rootHeight;
        public float variation;

        public Height(float rootHeightIn, float variationIn)
        {
            this.rootHeight = rootHeightIn;
            this.variation = variationIn;
        }

        public Biome.Height attenuate()
        {
            return new Biome.Height(this.rootHeight * 0.8F, this.variation * 0.6F);
        }
    }

    public static class RngSpawn extends RngItem
    {
        public Class <? extends EntityLiving > entityClass;
        public int minGroupCount;
        public int maxGroupCount;

        public RngSpawn(Class <? extends EntityLiving > entityclassIn, int weight, int groupCountMin, int groupCountMax)
        {
            super(weight);
            this.entityClass = entityclassIn;
            this.minGroupCount = groupCountMin;
            this.maxGroupCount = groupCountMax;
        }

//        public String toString()
//        {
//            return this.entityClass.getSimpleName() + "*(" + this.minGroupCount + "-" + this.maxGroupCount + "):" + this.weight;
//        }
    }

    public static enum TempCategory
    {
        SEA,
        COLD,
        MEDIUM,
        WARM;
    }
}
