package game.block;

import java.util.List;

import game.entity.Entity;
import game.entity.item.EntityTnt;
import game.entity.npc.EntityNPC;
import game.entity.projectile.EntityArrow;
import game.entity.types.EntityLiving;
import game.init.Items;
import game.init.SoundEvent;
import game.item.CheatTab;
import game.item.Item;
import game.item.ItemStack;
import game.material.Material;
import game.properties.IProperty;
import game.properties.PropertyBool;
import game.properties.PropertyInteger;
import game.renderer.blockmodel.ModelBlock;
import game.world.BlockPos;
import game.world.Explosion;
import game.world.Facing;
import game.world.State;
import game.world.World;
import game.world.WorldServer;

public class BlockTNT extends Block
{
    public static final PropertyBool EXPLODE = PropertyBool.create("explode");
    public static final PropertyInteger POWER = PropertyInteger.create("power", 0, 7);

    public BlockTNT()
    {
        super(Material.tnt);
        this.setDefaultState(this.getBaseState().withProperty(EXPLODE, Boolean.valueOf(false)));
        this.setTab(CheatTab.tabTech);
    }

    public void onBlockAdded(WorldServer worldIn, BlockPos pos, State state)
    {
        super.onBlockAdded(worldIn, pos, state);

        if (worldIn.isBlockPowered(pos))
        {
            this.onBlockDestroyedByPlayer(worldIn, pos, state.withProperty(EXPLODE, Boolean.valueOf(true)));
            worldIn.setBlockToAir(pos);
        }
    }

    /**
     * Called when a neighboring block changes.
     */
    public void onNeighborBlockChange(World worldIn, BlockPos pos, State state, Block neighborBlock)
    {
        if (worldIn.isBlockPowered(pos))
        {
            this.onBlockDestroyedByPlayer(worldIn, pos, state.withProperty(EXPLODE, Boolean.valueOf(true)));
            worldIn.setBlockToAir(pos);
        }
    }

    /**
     * Called when this Block is destroyed by an Explosion
     */
    public void onBlockDestroyedByExplosion(World worldIn, BlockPos pos, Explosion explosionIn, State prevState)
    {
        if (!worldIn.client)
        {
            EntityTnt entitytntprimed = new EntityTnt(worldIn, (double)((float)pos.getX() + 0.5F), (double)pos.getY(), (double)((float)pos.getZ() + 0.5F), explosionIn.getExplosivePlacedBy(), prevState.getValue(POWER).intValue());
            entitytntprimed.fuse = worldIn.rand.zrange(entitytntprimed.fuse / 4) + entitytntprimed.fuse / 8;
            worldIn.spawnEntityInWorld(entitytntprimed);
        }
    }

    /**
     * Called when a player destroys this Block
     */
    public void onBlockDestroyedByPlayer(World worldIn, BlockPos pos, State state)
    {
        this.explode(worldIn, pos, state, (EntityLiving)null);
    }

    public void explode(World worldIn, BlockPos pos, State state, EntityLiving igniter)
    {
        if (!worldIn.client)
        {
            if (((Boolean)state.getValue(EXPLODE)).booleanValue())
            {
                EntityTnt entitytntprimed = new EntityTnt(worldIn, (double)((float)pos.getX() + 0.5F), (double)pos.getY(), (double)((float)pos.getZ() + 0.5F), igniter, state.getValue(POWER).intValue());
                worldIn.spawnEntityInWorld(entitytntprimed);
                worldIn.playSoundAtEntity(entitytntprimed, SoundEvent.FUSE, 1.0F, 1.0F);
            }
        }
    }

    public boolean onBlockActivated(World worldIn, BlockPos pos, State state, EntityNPC playerIn, Facing side, float hitX, float hitY, float hitZ)
    {
        if (playerIn.getCurrentEquippedItem() != null)
        {
            Item item = playerIn.getCurrentEquippedItem().getItem();

            if (item == Items.flint_and_steel || item == Items.fire_charge)
            {
                this.explode(worldIn, pos, state.withProperty(EXPLODE, Boolean.valueOf(true)), playerIn);
                worldIn.setBlockToAir(pos);

                if (item == Items.flint_and_steel)
                {
                    playerIn.getCurrentEquippedItem().damageItem(1, playerIn);
                }
                else // if (!playerIn.creative)
                {
                    --playerIn.getCurrentEquippedItem().stackSize;
                }

                return true;
            }
        }

        return super.onBlockActivated(worldIn, pos, state, playerIn, side, hitX, hitY, hitZ);
    }

    /**
     * Called When an Entity Collided with the Block
     */
    public void onEntityCollidedWithBlock(World worldIn, BlockPos pos, State state, Entity entityIn)
    {
        if (!worldIn.client && entityIn instanceof EntityArrow)
        {
            EntityArrow entityarrow = (EntityArrow)entityIn;

            if (entityarrow.isBurning())
            {
                this.explode(worldIn, pos, worldIn.getState(pos).withProperty(EXPLODE, Boolean.valueOf(true)), entityarrow.shootingEntity instanceof EntityLiving ? (EntityLiving)entityarrow.shootingEntity : null);
                worldIn.setBlockToAir(pos);
            }
        }
    }
    
    public void getSubBlocks(Item itemIn, CheatTab tab, List<ItemStack> list)
    {
        for(int z = 0; z < 8; z++) {
        	list.add(new ItemStack(itemIn, 1, z));
        }
    }

    /**
     * Return whether this block can drop from an explosion.
     */
    public boolean canDropFromExplosion(Explosion explosionIn)
    {
        return false;
    }

    /**
     * Convert the given metadata into a BlockState for this Block
     */
    public State getStateFromMeta(int meta)
    {
        return this.getState().withProperty(EXPLODE, Boolean.valueOf((meta & 1) > 0)).withProperty(POWER, Integer.valueOf(((meta >> 1) & 7)));
    }

    /**
     * Convert the BlockState into the correct metadata value
     */
    public int getMetaFromState(State state)
    {
        return (((Boolean)state.getValue(EXPLODE)).booleanValue() ? 1 : 0) | (state.getValue(POWER).intValue() << 1);
    }
    
    public int damageDropped(State state)
    {
        return ((Integer)state.getValue(POWER)).intValue();
    }

    protected IProperty[] getProperties()
    {
        return new IProperty[] {EXPLODE, POWER};
    }
    
    public boolean isXrayVisible()
    {
        return true;
    }

    public ModelBlock getModel(String name, State state) {
    	String power = state.getValue(POWER) == 0 ? "" : ("_" + state.getValue(POWER));
    	return new ModelBlock("tnt_side" + power).add().nswe().d("tnt_bottom" + power).u("tnt_top" + power);
    }

    public IProperty<?>[] getIgnoredProperties() {
    	return new IProperty[] {EXPLODE};
    }

    public void onDestroyedByFire(World world, BlockPos pos, State state) {
    	this.onBlockDestroyedByPlayer(world, pos, state.withProperty(BlockTNT.EXPLODE, Boolean.valueOf(true)));
    }
}
